#==============================================================================
#
#  $Id$
#
"""
   Data model module.
"""
#
#  Copyright (C) 2005 Michael A. Muller
#
#  Permission is granted to use, modify and redistribute this code,
#  providing that the following conditions are met:
#
#  1) This copyright/licensing notice must remain intact.
#  2) If the code is modified and redistributed, the modifications must 
#  include documentation indicating that the code has been modified.
#  3) The author(s) of this code must be indemnified and held harmless
#  against any damage resulting from the use of this code.
#
#  This code comes with ABSOLUTELY NO WARRANTEE, not even the implied 
#  warrantee of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
#
#==============================================================================

# access modes
PUBLIC = 0
PRIVATE = 1
PROTECTED = 2

class SourceLocation(object):
   """
      Specifies a location in the source code.

      Public-vars:
         name::
            [string] source name
         lineNum::
            [int] line number
   """
   
   __slots__ = [ 'name', 'lineNum' ]

   def __init__(self, name, lineNum):
      self.name = name
      self.lineNum = lineNum

class Nested(object):
   """
      Class for objects that have a parent context.

      public-vars:
	 parent::
	    [Context or None] the parent context
   """

   __slots__ = [ 'parent' ]

   def __init__(self, parent):
      self.parent = parent
      
class Context(Nested):
   """
      Public-vars:
         names::
            [dict<string, Definition>] names scoped to the context.
   """
   
   def __init__(self, parent = None):
      super(Context, self).__init__(parent)
      self.names = {}
      self.seq = []

   def lookUp(self, name):
      """
         Looks up the name in the context and all parent contexts.  Returns
	 the definition if it exists, None if not.

	 parms:
	    name::
	       [string or list<string>] If this is a string, just looks up the
	       symbol.  If it is a list, does a recursive lookup (looks up
	       name 0 in the receiver, looks up name 1 in the result ...)
      """
      if isinstance(name, list):
	 # if this is a list, lookup the first part of the name then lookup
	 # the rest of the name in the result

	 defn = self.lookUp(name[0])
	 if len(name) > 1 and not isinstance(defn, Context):
	    raise ValueError(name)

	 # if this is a single-element name, we're done
	 if len(name) == 1:
	    return defn

	 try:
	    return defn.lookUp(name[1:])
	 except ValueError:
	    raise ValueError(name)
      else:
	 # simple name (just a string)

	 # try to get it locally
	 defn = self.names.get(name)

	 # if that fails, try the parent
	 if not defn and self.parent:
	    return self.parent.lookUp(name)
	 else:
	    return defn

   def getContents(self):
      """
         Returns all definitions in the context in _the order in which they
         were defined_.
      """
      return self.seq

   def append(self, item):
      """
	 Appends an item onto the sequence.  This should be used in cases
	 where the item is not a definition (like import statements, etc)
      """
      self.seq.append(item)

   def __setitem__(self, key, val):
      """
         Defines an item in the context.  The only twist is that it preserves
         the order of definition.
      """
      self.names[key] = val
      self.seq.append(val)

   def __getitem__(self, key):
      return self.names[key]

   def has_key(self, key):
      return self.names.has_key(key)

   def get(self, key):
      return self.names.get(key)

   def values(self):
      return self.names.values()

   def keys(self):
      return self.names.keys()

   def items(self):
      return self.names.items()

class Module(Context):
   pass
#   def __init__(self, name = None):
#      Context.__init__(self)
#      self.name = name
#
class Definition(Nested):
   """
      Base class for all definitions.  Definitions have a name and a source
      location - that being the source location of the name token.

      Public-vars::
         name::
            [string] the symbol name
         src::
            [@SourceLocation] where the definition lives in the source code
         access::
            [int] The access mode.  PROTECTED, PRIVATE or PUBLIC.
   """

   def __init__(self, name = None, src = None, parent = None, access = PUBLIC):
      super(Definition, self).__init__(parent)
      self.name = name
      self.src = src
      self.access = access
      self.__attrs = {}
      self.__fullNameVar = None

   def getTypeName(self):
      """
         Returns the definition type name (for use in error messages).  This
         defaults to returning the class name, but should be overriden by
         derived classes.
      """
      return self.__class__.__name__

   def getFullName(self):
      """
         Returns the fully-qualified name of the definition as a string.
      """
      return '::'.join(self.fullName)

   def setAttr(self, name, val):
      """
         Sets an attribute.  Attributes are used by annotations to
         store arbitrary information (most likely the annotation object
         itself).

         parms:
            name::
               [tuple<string, string>] fully qualified attribute name.
               Attribute names consist of an attribute namespace followed by a
               name.  The namespace "cppre" is reserved and used for all of
               the standard system attributes.  All user defined namespaces
               should be in the form of a URI.
            val::
               [any] the value associated with the attribute
      """
      self.__attrs[name] = val

   def getAttr(self, name):
      """
         Returns the value of an attribute, *None* if the attribute is not
         defined.

         parms:
            [tuple<string, string>] attribute namespace and name
      """
      return self.__attrs.get(name)

   def __fullName(self):
      if not self.__fullNameVar:
	 if self.parent and isinstance(self.parent, Definition):
	    parentName = self.parent.fullName
	 else:
	    parentName = []
	 self.__fullNameVar = parentName + [self.name]

      return self.__fullNameVar

   fullName = property(__fullName)

class Namespace(Definition, Context):

   def __init__(self, name = None, src = None, parent = None):
      super(Namespace, self).__init__(name, src, parent)

   def getTypeName(self):
      return 'namespace'

class Class(Definition, Context):
   
   def __init__(self, name = None, src = None, parent = None, bases = None):
      super(Class, self).__init__(name, src, parent)
      self.bases = bases

   def toSourceForm(self):
      return self.getFullName()

   def getTypeName(self):
      return 'class'

class Global(Context):
   """
      The global context.  Contains all of the primitive types and all
      modules.
   """

   intType = Class('int')
   charType = Class('char')
   floatType = Class('float')
   doubleType = Class('double')
   stringType = Class('string')
   voidType = Class('void')

   def __init__(self):
      super(Global, self).__init__()

      # add the standard types to the module
      self['int'] = self.intType
      self['char'] = self.charType
      self['float'] = self.floatType
      self['double'] = self.doubleType
      self['string'] = self.stringType
      self['void'] = self.voidType
      
class TypeSpec(object):
   """
      A type specification.

      Public-vars:
         const::
            [boolean]
         type::
            [@Class] the base type
   """

   def __init__(self):
      super(TypeSpec, self).__init__()
      self.const = False
      self.type = None

   def toSourceForm(self):
      """
         Returns the source representation of the type spec as a string.
      """
      val = ''
      if self.const:
         val += 'const '

      val += self.type.toSourceForm()
      return val

class VarDef(Definition):
   """
      A variable definition.

      Public-vars:
         type::
            [@TypeSpec] the variable type
   """

   def __init__(self, name = None, src = None, parent = None, type = None):
      super(VarDef, self).__init__(name, src)
      self.type = type

   def getTypeName(self):
      return 'variable'

class FunctionDef(Definition):
   """
      A function or method definition.

      Public-vars:
         returnType::
            [@TypeSpec] the function return type
         parms::
            [list<@VarDef>] the function's formal arguments.
         const::
            [boolean] true if this is a const method
         body::
            [@FunctionBody or None] the function body
   """

   def __init__(self, name = None, src = None, parent = None, body = None):
      super(FunctionDef, self).__init__(name, src, parent)
      self.returnType = None
      self.parms = []
      self.const = False
      self.body = None

   def getTypeName(self):
      return 'function'

class FunctionBody(object):
   """
      A function body.  Currently we preserve this as a list of tokens.
   """

   __slots__ = [ '_FunctionBody__tokens' ]

   def __init__(self, tokens):
      super(FunctionBody, self).__init__()
      self.__tokens = tokens

   def generateCpp(self, out):
      """
         Generates the C++ code for the function body.  In the current
         implementation, this merely replays the token list onto the  output
         stream.
      """
      if self.__tokens:
         # output the source line number (really need to do this for every
         # token where there is a discontinuity)
         tok = self.__tokens[0]
         out.write('#line %d "%s"\n' % (tok.src.lineNum, tok.src.name))

         # write the tokens
         for tok in self.__tokens:
            out.write(tok.val)

class InappropriateAnnotationError(Exception):
   """
      Raised when an attempt is made to apply an annotation to a type of
      definition that it doesn't support.
   """
   pass

class Annotation(object):
   """
      Base class for all annotations.

      Annotations are similar to annotations in Java (or "decorators" in
      Python).  They allow you to modify the meta-data for the definition that
      follows them.

      Public-vars:
         name::
            [string] the annotation name
   """

   __slots__ = [ 'name' ]

   def __init__(self, name):
      super(Annotation, self).__init__()
      self.name = name

   def applyTo(self, defn):
      """
         Applies the annotation to the given definition.  Should verify that
         the annotation is appropriate for the definition.  It may modify the
         definition or associate itself with it.

         Must be implemented by derived classes.

         parms:
            defn::
               [@Definition]
      """
      raise NotImplementedError()

class AttrAnn(Annotation):
   """
      Annotation which promotes an instance variable definition to an
      "attribute" - a variable with a getter and setter.

      When applied to a variable definition, this sets the attribute 
      "cppre:attr" to the instance.
   """

   def applyTo(self, defn):
      if not isinstance(defn, VarDef):
         raise InappropriateAnnotationError()
      
      defn.setAttr(('cppre', 'attr'), self)

class IncludeStmt(object):
   """
      Public-vars::
         src::
            [@SourceLocation] source location of the include statement
         file::
            [string] the filename to be included
   """

   __slots__ = [ 'src', 'file' ]

   def __init__(self, file = None, src = None):
      super(IncludeStmt, self).__init__()
      self.src = src
      self.file = file

class ImportStmt(object):
   """
      Public-vars::
	 src::
	    [@SourceLocation] source location of the import statement
	 name::
	    [list<string>] fully qualifed name
   """

   __slots__ = [ 'src', 'name' ]

   def __init__(self, name = None, src = None):
      super(ImportStmt, self).__init__()
      self.src = src
      self.name = name

   def createDefinitions(self, context):
      """
	 Creates the definition tree corresponding to the import statement in
	 the given context.

	 Raises a TypeError if one of the non-final names exists and is not a
	 namespace, or the final name exists and is not a class.

	 parms:
	    context::
	       [@Context]
      """
      def createDefs(context, name):

	 # get the first name component
	 nameComp = name[0]

	 if len(name) > 1:
	    nested = context.lookUp(nameComp)

	    # create a new namespace if it doesn't exist
	    if not nested:
	       nested = Namespace(nameComp, self.src, context)
	       context[nameComp] = nested

	    # verify that it is a namespace
	    elif not isinstance(nested, Namespace):
	       raise TypeError(nameComp + ' is not a namespace')

	    # recurse into it
	    createDefs(nested, name[1:])

	 else:
	    # make sure that if it exists, it is a class
	    nested = context.lookUp(nameComp)
	    if not nested:
	       nested = Class(nameComp, self.src, context)
	       context[nameComp] = nested
	    elif not isinstance(nested, Class):
	       raise TypeError(nameComp + ' is not a class')

      createDefs(context, self.name)

class UsingStmt(object):
   """
      Public-vars:
	 src::
	    [@SourceLocation] source location of the using statement
	 namespace::
	    [@Namespace] the namespace to use all of.
   """

   __slots__ = [ 'namespace', 'src' ]

   def __init__(self, namespace, src = None):
      self.namespace = namespace
      self.src = src
