#==============================================================================
#
#  $Id$
#
"""
   CPPre parser.
"""
#
#  Copyright (C) 2005 Michael A. Muller
#
#  Permission is granted to use, modify and redistribute this code,
#  providing that the following conditions are met:
#
#  1) This copyright/licensing notice must remain intact.
#  2) If the code is modified and redistributed, the modifications must 
#  include documentation indicating that the code has been modified.
#  3) The author(s) of this code must be indemnified and held harmless
#  against any damage resulting from the use of this code.
#
#  This code comes with ABSOLUTELY NO WARRANTEE, not even the implied 
#  warrantee of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
#
#==============================================================================

from tok import CToken, Tokenizer
from model import *

class ParseError(Exception):
   """
      Raised whenever a parse error (syntax error) occurs.
   """
   def __init__(self, tok, msg, **kw):
      Exception.__init__(self, tok, msg)
      self.tok = tok
      self.msg = msg
      self.vars = kw

   def __str__(self):
      # make a variable dictionary for the message string from the keyword
      # args and the standard variables.
      vars = self.vars.copy()
      vars.update({'token': self.tok.val})

      return '%s:%d: %s' % \
         (self.tok.src.name, self.tok.src.lineNum, 
          self.msg % vars
          )

class Parser:
   """
      The parser class.  To use it:

      {{
         src = open('filename.cpr')
         parser = Parser(src, 'filename.cpr')
         mdoule = parser.parse()
      }}
   """

   def __init__(self, globalContext, src, sourceName, lineNum = 1):
      self.__toker = Tokenizer(src, CToken.definition)

      # the context stack - put the global context on it
      self.__contexts = [globalContext]

      # create a context for the module, make it the current context
      self.__push(Module(globalContext))

      # the annotations dictionary.  A mapping from annotation names to
      # callable objects that can be used to construct instances of those
      # annotations.
      self.__annotations = {
         'attr': AttrAnn,
      }

   def defineAnnotation(self, name, annotation):
      """
         Defines an annotation.

         parms:
            name::
               [string]
            annotation::
               [class] class derived from @model.Annotation
      """
      self.__annotations[name] = annotation

   def __push(self, context):
      """
         Push a new context onto the stack.
      """
      self.__context = context
      self.__contexts.append(context)

   def __pop(self):
      """
         Pops the top context and returns it.
      """
      ctx = self.__context
      self.__contexts.pop()
      self.__context = self.__contexts[-1]
      return ctx

   def parseToken(self, newlineIsSemicolon = False):
      """
         Returns the next non-whitespace, non-comment token.

	 parms:
	    newlineIsSemicolon::
	       [boolean] if true, converts a newline to a semicolon
      """
      tok = CToken(CToken.whitespace, '', SourceLocation('', 0))
      while tok.isSkipable():
         tok = self.__toker.getToken()

	 # do semicolon substitution
	 if newlineIsSemicolon and tok.isWhitespace() and '\n' in tok.val:
	    tok.type = CToken.symbol
	    tok.val = ';'

      return tok

   def parse(self):
      """
         Parse the outer layer.  Returns a @model.Module instance.
      """
      tok = self.parseToken()

      # set the default module name
      self.__context.name = 'default'

      annotations = []

      while not tok.isEnd():
         if tok.isIdent():
            if tok.val == 'class':
               cls = self.parseClassDef()
               for ann in annotations:
                  ann.applyTo(cls)
               annotations = []
	    elif tok.val == 'import':
	       self.parseImportStmt()
            elif tok.val == 'include':
               self.parseIncludeStmt()
	    elif tok.val == 'using':
	       self.parseUsingStmt()
            elif tok.val == 'module':
               self.__context.name = self.parseModuleStmt()
            else:
               self.__toker.putBack(tok)
               defn = self.parseDefinition()
               for ann in annotations:
                  ann.applyTo(defn)
               annotations = []
               self.__define(defn)
         elif tok.isSymbol('@'):
            annotations.append(self.parseAnnotation())
         elif tok.isEnd():
            return
         else:
            raise ParseError(tok, '"%(token)s was not expected at this time')

         # get the next token
         tok = self.parseToken()

      return self.__context

   def parseSymbol(self, symbol):
      """
         Parse the next token and throw a ParseError if it is not the given
         symbol.

         parms:
            symbol::
               [string] the symbol to check for.
      """
      tok = self.parseToken()
      if not tok.isSymbol(symbol):
         raise ParseError(tok, 'Expected "%(sym)s", not "%(token)s"',
                          sym = symbol
                          )

   def parseIdent(self):
      """
         Parse an identifier, return the token.
      """
      id = self.parseToken()
      if not id.isIdent():
         raise ParseError(id, 'Identifier expected, not "%(token)s"')
      return id

   def parseNamespace(self, newlineTerminated):
      """
	 Parses a fully-qualified namespace name (verifying that it is a
	 namespace) and returns the @Namespace instance for it.

	 parms:
	    newlineTerminated::
	       [boolean] set to true in a context where the namespace is the
	       last item in a statement that can be terminated by a newline
      """
      name = []
      while True:
	 # parse the identifier
	 nameTok = self.parseIdent()
	 name.append(nameTok.val)

	 # get the next token, check for a colon
	 tok = self.parseToken(newlineIsSemicolon = newlineTerminated)
	 if not tok.isSymbol(':'):
	    self.__toker.putBack(tok)
	    break

	 # check for another colon, rollback both if we didn't get it
	 tok2 = self.parseToken()
	 if not tok2.isSymbol(':'):
	    self.__toker.putBack(tok2)
	    self.__toker.putBack(tok)
	    break

      # look up the namespace
      try:
	 namespace = self.__context.lookUp(name)
	 if not isinstance(namespace, Namespace):
	    raise ValueError(name)
      except ValueError:
	 raise ParseError(nameTok, '%s is not a namespace' % '::'.join(name))

      return namespace
      
   def parseTypeSpec(self):
      """
         Parses a full type spec and returns it (a @TypeSpec instance)
      """
      typeInfo = TypeSpec()
      typeInfo.const = False

      tok = self.parseIdent()

      # check for the modifiers
      gotConst = gotStatic = False
      while True:
         if tok.val == 'const':
            if gotConst:
               raise ParseError(tok, 'duplicate "const"')
            typeInfo.const = True
            tok = self.parseIdent()
         elif tok.val == 'static':
            if gotStatic:
               raise ParseError(tok, 'duplicate "static"')
            typeInfo.static = True
            tok = self.parseIdent()
         else:
            break

      # try to look up the type name
      type = self.__context.lookUp(tok.val)
      if not type:
         raise ParseError(tok, 'Unknown type: "%(token)s"')
      else:
         typeInfo.type = type

      return typeInfo

   def parseOperator(self):
      """
         Parse an operator overloading defintiion operator.  Returns a string
         containing a simplified version of the operator.

         {{
            operator * (
                    ^ ^
         }}
      """
      tok = self.parseToken()
      if tok.isSymbol() and tok.val in ('*', '+', '/'):
         return tok.val
      else:
         raise ParseError(tok, '%(token)s was unexpected at this time')

   def parseParmDef(self):
      """
         Parse a formal argument definition.  Returns a @VarDef instance.

         {{
            function( Type val, ...
                     ^         ^
         }}
      """
      typeSpec = self.parseTypeSpec()

      # parse the variable name
      varName = self.parseIdent()

      # get the next token
      tok = self.parseToken()

      if tok.isSymbol('='):
         raise ParseError(tok, 'XXX default values not supported')
      elif tok.isSymbol(','):
         pass
      elif tok.isSymbol(')'):
         self.__toker.putBack(tok)
      # XXX am thinking that self.__context is the class context but needs to
      # ultimately be the method context
      return VarDef(varName.val, varName.src, self.__context, typeSpec)

   def parseParmList(self):
      """
         Parse a formal argument list.  Returns a list<@VarDef> of the
         results.

         {{
            function( ... )
                     ^     ^
         }}
      """
      parms = []
      while True:
         # check for closing paren
         tok = self.parseToken()
         if tok.isSymbol(')'):
            break

         # parse a variable definition
         self.__toker.putBack(tok)
         parms.append(self.parseParmDef())

      return parms

   def parseFunctionBody(self):
      """
         Parse the body of a function.  Returns a @FunctionBody instance.

         function(...) { ... }
                      ^       ^
      """
      tokens = []

      # first token must be a bracket
      tok = self.__toker.getToken()
      if not tok.isSymbol('{'):
         raise ParseError(tok, 'Expected "{", got "%(token)s"')
      tokens.append(tok)

      # parse and preserve until the bracket count drops to zero
      bracketCount = 1
      while bracketCount:
         tok = self.__toker.getToken()
         if tok.isSymbol('}'):
            bracketCount -= 1
         elif tok.isSymbol('{'):
            bracketCount += 1

         tokens.append(tok)

      return FunctionBody(tokens)

   def parseDefinition(self):
      """
         Parse a definition and return it.
      """
      typeSpec = self.parseTypeSpec()

      # parse the variable name
      varName = self.parseIdent()
      varNameVal = varName.val
      startPos = varName.src

      # if this is an operator overload, get the operator
      isOperator = False
      if varName.isIdent('operator'):
         varNameVal = self.parseOperator()
         isOperator = True

      # get the next token
      tok = self.parseToken()

      if tok.isSymbol('('):
         # function definition
         defn = FunctionDef(varNameVal, startPos, parent = self.__context)
         defn.returnType = typeSpec
         defn.parms = self.parseParmList()

         tok = self.parseToken()

         # check for "const"
         if tok.isIdent('const'):
            if not isinstance(self.__context, Class):
               raise ParseError(tok, 'Only class methods may be "const"')

            defn.const = True
            tok = self.parseToken()

         if tok.isSymbol('{'):
            self.__toker.putBack(tok)
            defn.body = self.parseFunctionBody()
            return defn
         elif tok.isSymbol(';'):
            return defn
         else:
            raise ParseError(tok, '%(token)s not expected at this time')

      # complain if this was an operator definition and we didn't get the
      # start of an argument list
      elif isOperator:
         raise ParseException('"(" expected, not %(token)s')

      # check for initialization
      elif tok.isSymbol('='):
         raise ParseError(tok, 'XXX initialization not supported yet')

      # check for more definitions
      elif tok.isSymbol(','):
         raise ParseError(tok, 'XXX multiple definitions not supported yet')

      elif not tok.isSymbol(';'):
         raise ParseError(tok, '%(token)s not expected here')

      return VarDef(varNameVal, startPos, self.__context, typeSpec)

   def parseAnnotation(self):
      """
         Parse an annotation.

         {{
            @ annotation
             ^          ^
         }}
      """
      # get the annotation name, make sure it has been defined
      name = self.parseIdent()
      try:
         annotation = self.__annotations[name.val](name.val)
      except KeyError:
         raise ParseError(name, 'Undefined annotation: %(token)s')

      return annotation

   def parseAccessModifier(self, tok, mode, annotations):
      """
         Parses an access modifier (public, private, protected) and its colon.
         Also verifies that this is appropriate.

         {{
            public :
                  ^ ^
         }}
      """
      # verify that we don't have any annotations
      if annotations:
         raise ParseError(tok, 
                          'Annotations must preceed a definition: found '
                           'an access modifier'
                          )
      self.parseSymbol(':')
      return mode

   def parseClassBody(self, cls):
      """
         Parse the class body from after the start curly-brace.
      """
      mode = PRIVATE
      annotations = []
      while True:
         tok = self.parseToken()
         if tok.isSymbol('}'):
            return
         elif tok.isSymbol('@'):
            # got the magic at sign - parse an annotation
            annotations.append(self.parseAnnotation())
         elif tok.isIdent():

            # check for all the reserved words
            if tok.val == 'public':
               mode = self.parseAccessModifier(tok, PUBLIC, annotations)
            elif tok.val == 'private':
               mode = self.parseAccessModifier(tok, PRIVATE, annotations)
            elif tok.val == 'protected':
               mode = self.parseAccessModifier(tok, PROTECTED, annotations)
            else:
               # otherwise it is a definition
               self.__toker.putBack(tok)
               defn = self.parseDefinition()

               # store the access mode
               defn.access = mode

               # apply all annotations
               for ann in annotations:
                  try:
                     ann.applyTo(defn)
                  except InappropriateAnnotationError:
                     raise ParseError(tok, 
                                      'The %(ann)s annotation can not be '
                                       'applied to a %(defType)s', 
                                      ann = ann.name, 
                                      defType = defn.getTypeName()
                                      )
               annotations = []

               # store it in the current context
               self.__define(defn)

   def __define(self, defn):
      # verify that the name has not already been defined in this context.
      if self.__context.has_key(defn.name):
         other = self.__context[defn.name]
         raise ParseError(tok, 
                          'name %(token)s was already defined at '
                           '%(src)s:%(lineNum)d', 
                          src = other.src.name,
                          lineNum = other.src.lineNum
                          )
      self.__context[defn.name] = defn

   def parseClassDef(self):
      """
         Parse a class definition from after the "class" keyword.  returns the
         class context.
      """
      # create the class object
      cls = Class()

      # parse the class name
      tok = self.parseIdent()
      cls.name = tok.val
      cls.src = tok.src

      # add the class to the current context and then make it the current
      # context
      self.__define(cls)
      cls.parent = self.__context
      self.__push(cls)

      baseClasses = []

      # get the next token
      tok = self.parseToken()
      
      # check for base class definitions
      if tok.isSymbol(':'):
         baseClasses = self.parseBaseClasses()
         tok = self.parseToken()

      cls.bases = baseClasses

      if tok.isSymbol('{'):
         body = self.parseClassBody(cls)
      else:
         raise ParseError(tok, '"%(token)s was unexpected at this time')

      # pop the class context
      self.__pop()

      return cls

   def parseImportStmt(self):
      name = []
      tok = self.parseIdent()
      name.append(tok.val)
      startLoc = tok.src

      while True:
	 tok = self.parseToken(newlineIsSemicolon = True)
	 if tok.isSymbol(';'):
	    break

	 # parse the first colon
	 if not tok.isSymbol(':'):
	    raise ParseError(tok, '"::" or end-statement expected in import')

	 # parse the second colon
	 self.parseSymbol(':')

	 tok = self.parseIdent()
	 name.append(tok.val)

      # create the import statement and add it to the sequence
      imp = ImportStmt(name, startLoc)
      self.__context.append(imp)

      # create the corresponding hierarchy of namespaces and classes
      imp.createDefinitions(self.__context)

   def parseIncludeStmt(self):
      tok = self.parseToken()
      if not tok.isString():
         raise ParseError(tok, 'String expected, not %(token)s')

      self.parseSymbol(';')
      self.__context.append(IncludeStmt(eval(tok.val), tok.src))

   def parseUsingStmt(self):
      tok = self.parseIdent()
      if tok.val != 'namespace':
	 raise ParseError(tok, 
			  'Can only deal with "using namespace" at this time'
			  )
      startLoc = tok.src

      # parse the namespace name and get the namespace object
      ns = self.parseNamespace(True)

      # parse the trailing semicolon or newline
      tok = self.parseToken(newlineIsSemicolon = True)
      if not tok.isSymbol(';') and not tok.isEnd():
	 raise ParseError(tok, 'newline or semicolon expected')

      # add the using statement to the sequence
      self.__context.append(UsingStmt(ns, startLoc))

      # import all of the objects in the namespace into the current context
      for name, item in ns.items():
	 if self.__context.has_key(name):
	    raise ParseError(tok, 
			     'Name conflict for %s from namespace %s' %
			      (name, ns.fullName)
			     )

	 self.__context[name] = item

   def parseModuleStmt(self):
      tok = self.parseIdent()
      self.parseSymbol(';')
      return tok.val

