#!/usr/local/bin/python
#==============================================================================
#
#  $Id: tok.py 842 2004-05-04 01:54:45Z mike $
#
"""
   Simple tokenizer framework.
   
   At this time, all of the functionality of the @CTokenizer is
   implemented in @Tokenizer.  Eventually, the *CTokenizer* personality
   should be broken out into the correct class.
   
   Synopsis:
   
   {{

      from spug.util.tok import CTokenizer
      
      # create a new tokenizer to parse standard input
      toker = CTokenizer(sys.stdin)
      
      while 1:
      
         # get the next token
         tok = toker.nextToken()
         
         # check for end-of-stream
         if tok.isType(Token.end):
            break
         
         # print out the token type and the value
         print tok.type.name + ': ' + tok.val
   }}
   
   Run "#python tok.py#" to see the action of the above code.
      
""" #"
#
#  Copyright (C) 1998, 2006 Michael A. Muller
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#  $Log: tok.py,v $
#  Revision 1.4  2004/05/04 01:54:45  mike
#  Misc small fixes, added a TokenStream class to avoid having to do that with
#  every new client.
#
#  Revision 1.3  2000/02/13 21:58:52  mike
#  Doc fixes.
#
#  Revision 1.2  2000/02/13 00:09:08  mike
#  Converted comments to docstrings.
#
#  Revision 1.1  1998/10/22 23:46:26  mike
#  Initial revision
#
#
#==============================================================================

import re

class SourceLocation:
   """
      Represents the source location of a token.

      Public-vars:
         name::
            [string] source file name
         lineNum::
            [ing] source line number
   """

   def __init__(self, name, lineNum):
      self.name = name
      self.lineNum = lineNum

class Token:
   """
      Tokens represent pieces of text.  This is the basic definition of a
      token.  Tokenizer implementations are free to extend this to add
      functionality.

      Public-vars:
         val:: 
            source text of the token.  Its "value".
         type:: 
            a numeric value indicating the tokens type
         src::
            [@SourceLocation] the source location that we got the token from
   """

   def __init__(self, type, val, src):
      """
         Constructor for Token.  Type is one of the types listed above, val is
         the value of the token (its text), srcName is the name of the source
         file that the token was tokenized from, and /lineNum/ is the line
         number in the source file.

         /srcInfo/ is a tuple indicating the source file name and the line
         number.
      """
      
      self.type = type
      self.val = val
      self.src = src

   def isType(self, type):
      """
         Returns true if the token is of the indicated type.
      """
      return self.type is type

   def equals(self, type, val):
      """
         Returns true if the token is of the indicated type and has the
         indicated value.
      """
      return self.type == type and self.val == val
   
class TokenInfo:
   """
      The formal definition of a token type.

      Public-vars:
         name::
            the name of the token type
         regex::
            the regular expression that describes the tokens source form
         tokenFactory::
             [callable<@TypeInfo, string, @SourceLocation>] a function that
             constructs and returns a token.
         continued::
            an optional regular expression.  If it is present, it indicates
            that the token may be continued over multiple lines (if /regex/
            matches to the end of the current line) and it represents the kind
            of expression which will terminate the multi line token beginning
            with /regex/.  All lines between the line that begins the token
            and the portion of a line which ends the token are considered to
            be part of the token.
   """

   def __init__(self, name, regex, tokenFactory, continued):
      """

      """

      self.name = name
      self.regex = regex
      self.tokenFactory = tokenFactory
      self.continued = continued

   def create(self, rawVal, srcLoc):
      return self.tokenFactory(self, rawVal, srcLoc)

class TokenizerError(Exception):
   "A Tokenizer error"
   
   def __init__(self, toker, text):

      Exception.__init__(self, 
                         toker.src.name + ":" + `toker.lineNum` + ": " + text
                         )    
      self.toker = toker

class TokenizerDefinition:
   """
      This is the abstract interface that must be defined to implement a
      tokenizer.  There are only two requirements:

      -  you must be able to iterate over it.  Iteration should produce
         @TokenInfo instances in the order that they are to be checked against
         the current stream position.
      -  you must provide a function to create the end token.
   """

   def __iter__(self):
      raise NotImplementedError()

   def createEndToken(self):
      """
         Returns a new end token (indicating the end of the input stream).

         Must be implemented by derived classes.
      """
      raise NotImplementedError()

class Tokenizer:
   """
      A Tokenizer is used to extract tokens from a source stream.  Tokens
      are of the form normally accepted by C-like languages.
   """
   
   def __init__(self, src, tokenizerDef, lineNum = 0):
      """
         Constructor.  Creates a new Tokenizer given a source file and
         (optionally) a set of flags.
      """

      self.src = src
      self.lineNum = lineNum
      self.preparsed = [ ]
      self.buffer = ""
      self.tokerDef = tokenizerDef
   
   def getNextPreparsed(self):
      """
         Return the next token in the preparsed list (the list of tokens that
         has already been parsed and were put back).
      """
      
      next = self.preparsed[0]
      del self.preparsed[0]
      return next
   
   def fillBuffer(self):
      """
         Make sure that the buffer has data in it.
      """
      
      if not self.buffer:
         self.buffer = self.src.readline()
         self.lineNum = self.lineNum + 1

   def breakOff(self, toksrc):
      """
         "breaks off" the regular expression match specified by toksrc from the
         buffer, and returns it.
      """
   
      self.buffer = self.buffer[toksrc.end():]
      return toksrc.group()

   def __takeIt(self, tokType, toksrc):
      """
         Takes the next token off of the front of the buffer and returns it.
	 If the token type continues onto the next line, continue it.
      """
      
      # get the stuff
      stuff = self.breakOff(toksrc)
      
      # if the token continues onto the next line and the regular
      # expression matches rest of the current line, continue refilling
      # the buffer until we obtain an expression that matches the
      # "continued" expression.
      if tokType.continued and not len(self.buffer):
         
#         print "continued is regex: ",  \
#            isinstance(tokType.continued, re.RegexObject)
         
         self.fillBuffer()
         
         # keep getting until we either find the terminator or run out
         # of source code
         while self.buffer:
         
            # quit if we find the end
            toksrc = tokType.continued.match(self.buffer)
            if toksrc:
               return stuff + self.breakOff(toksrc)
            
            # if not, take whatcha got and gimme more
            stuff = stuff + self.buffer
            self.buffer = None
            self.fillBuffer()
            
         else:
            raise TokenizerError(self, tokType.name + " never terminated.")
         
         return stuff
      else:
         return stuff

   def parseNextToken(self):
      """
         Parses the next token directly off of the stream.  Clients should
         generally avoid using this.  Use nextToken() instead, since that will
         use the preparsed queue if tokens have been put back.
      """
      
      # make sure that the buffer has data
      self.fillBuffer();

      # stash the source location
      srcLoc = SourceLocation(self.src.name, self.lineNum)
      
      # if the buffer is still empty, this is the end of the stream
      if not self.buffer:
         return self.tokerDef.createEndToken(srcLoc)
      
      # check each of the regular expressions against the contents of the
      # buffer
      for x in self.tokerDef:
      
         # see if this expression matches
         toksrc = x.regex.match(self.buffer) 
         if toksrc:
            
            # break the token off of the front of the stream
            toksrc = self.__takeIt(x, toksrc)
            
            # construct a new instance of x
            return x.create(toksrc, srcLoc)
            
      else:
         raise TokenizerError(self, "Unknown token.");
   
   def getToken(self):
      """
         Returns the next token, either from the preparsed cache, or from the
         stream.
      """
      
      if self.preparsed:
         return self.getNextPreparsed()
      else:
         return self.parseNextToken()

   def putBack(self, tok):
      """
         Puts the given token back on the list.  Puts it first on the list,
         so immediately after calling, /tok/ will be the next token
         return from @nextToken().
      """
      self.preparsed.insert(0, tok) 

   def __iter__(self):
      """
         Iterator method that returns self so that we can serve as our own
         iterator.
      """
      return self

   # alias the getToken() method so that we can act as an interator
   next = getToken

class CToken(Token):
   "Token set for C-like languages"

   # we don't appear to need to define this as staticmethod - the fact that
   # we're using it in the context of the class seems to be sufficient.
   def __create(type, raw, src):
      return CToken(type, raw, src)

   end = TokenInfo('end', None, __create, None)

   identifier = \
      TokenInfo("identifier", re.compile(r'[A-Za-z_][A-Za-z_0-9]*'), __create, 
                None
                )
      
   integer = \
      TokenInfo("integer", re.compile(r'(0x[0-9a-fA-F]+)|([0-9]+)'), __create,
                None
                )
      
   string = \
      TokenInfo("string", re.compile(r'"([^\\]|(\\.))*?"'), __create,
                None
                )
   
   character = \
      TokenInfo("character", re.compile(r" ' [^\\] | (\\.) ' ", re.X), 
                __create,
                None
                )
   
   # comments can consist of either // to end of line or /* */
   comment = \
      TokenInfo("comment", 
                re.compile(r' ( /\* .*?  \*/ ) |'
                            r'( // .* $ )', 
                           re.X
                           ),
                __create,
                None
                )

   # long comments are simply comments that extend past the end of the
   # line.
   longComment = \
      TokenInfo("longComment", re.compile(r' /\* .* \n$ ', re.X), __create,
                re.compile(r' .* \*/ ', re.X)
                )         
   
   whitespace = \
      TokenInfo("whitespace", re.compile(r'[ \n\r\t]+'), __create, None)

   symbol = \
      TokenInfo("symbol",
                re.compile(r'[@~!#%^&\*\(\)\-\+\=\|\}\{\]\[:;<>\?,\./\\]'),
                __create,
                None
                )

   __tokenTypes = [ 
      identifier, integer, string, comment, longComment, whitespace, symbol
   ]

   class _Definition(TokenizerDefinition):
      """
         The tokenizer definition for a C Tokenizer.
      """

      def __init__(self, tokenTypes, end):
         self.__tokenTypes = tokenTypes
         self.__endType = end

      def __iter__(self):
         return iter(self.__tokenTypes)

      def createEndToken(self, srcLoc):
         return CToken(self.__endType, '', srcLoc)

   definition = _Definition(__tokenTypes, end)

   def isSymbol(self, symVal = None):
      """
         Returns true if the token is a symbol whose text is /symVal/.
         If symVal is not specified, returns true if the token is a symbol.

         parms:
            symVal::
               [string or None] text of the symbol for comparison.
      """
      return self.type is self.symbol and \
         (symVal is None or self.val == symVal)

   def isString(self):
      "Returns true if the token is a string."
      return self.type is self.string

   def isInteger(self):
      "returns true if the token is an integer"
      return self.type is self.integer

   def isSkipable(self):
      "Returns true if the token is whitespace or a comment"
      return self.type in (self.whitespace, self.comment, self.longComment)

   def isWhitespace(self):
      "Returns true if the token is whitespace."
      return self.type is self.whitespace

   def isComment(self):
      "Returns true if the token is a comment"
      return self.type in (self.comment, self.longComment)

   def isIdent(self, val = None):
      """
         Returns true if the token is an identifier and the text matches /val/.
      """
      return self.type is self.identifier and (not val or val == self.val)

   def isEnd(self):
      "Returns true if the token is the end-of-stream token"
      return self.type is self.end

class TokenStream:
   """
      This class can be used as a wrapper for any object that provides
      the file readline() method - it delegates that and also provides
      a "name" variable, required by the tokenizer.
   """

   def __init__(self, src, name):
      self.src = src
      self.name = name

   def readline(self):
      return self.src.readline()

if __name__ == '__main__':
   import sys  
   toker = Tokenizer(sys.stdin, CToken.definition)
   
   while 1:
   
      tok = toker.getToken()
      if tok.isEnd():
         break
      
      print tok.type.name + ': ' + tok.val
