# XXX note - this module doesn't belong here.  Keeping it here for convenience
#==============================================================================
#
#  $Id: tester.py,v 1.3 2004/05/19 01:25:38 mike Exp $
#
"""
   Classes to facilitate automatic testing of code.
"""
#
#  Copyright (C) 2002 Michael A. Muller
#
#  Permission is granted to use, modify and redistribute this code,
#  providing that the following conditions are met:
#
#  1) This copyright/licensing notice must remain intact.
#  2) If the code is modified and redistributed, the modifications must 
#  include documentation indicating that the code has been modified.
#  3) The author(s) of this code must be indemnified and held harmless
#  against any damage resulting from the use of this code.
#
#  This code comes with ABSOLUTELY NO WARRANTEE, not even the implied 
#  warrantee of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
#
#==============================================================================

import sys, traceback

class BadTestCase(Exception):
   pass

class TestCase:
   
   """
      Public-vars::
         desc::
            [string] brief description of the test, suitable for inclusing
            in the string "testing <desc>...".  This is only used by the 
            @getDesc() method, and the test case programmer may opt to 
            override @getDesc() instead.
         deps::
            [list<string>] list of names of test cases that must be run
            prior to this test case.
         varsUsed::
            [dict<string, string or tuple<string, string>>] dictionary
	    mapping variable names to the name of the test case where those
            variables originated.  If the variable has a local name that is
            different from the name used in the originating test case, the
            associated value is a tuple defining the originating test case
            and the name of the variable in the originating test case.
         varsDefined::
            [list<string>] list of variables defined by this test case.
   """

   deps = []         
   varsUsed = {}
   varsDefined = []
      
   
   def run(self):
      """
         This is the heart of the test case.  This function should contain
         the actual test, returning true if the test passes, false if it 
         fails.
         
         Must be implemented by derived classes.
         
         parms:
            testbed::
               [@Testbed] testbed object from which to retrieve and in which
               to deposit computed values.
      """
      raise NotImplementedError()

   def log(self, msg):
      """
         Writes a message to the test log.
      
         parms:
            msg::
               [string]
      """
      getTestbed().log('%s: %s' % (self.getId(), msg))

   def getDesc(self):
      """
         Returns the test description (a string).
         
         May be implemented by derived classes.
      """
      return self.desc
   
   def getId(self):
      """
         Returns the test id (a string).  This is normally the test class name,
         but can be any name that uniquely identifies the test.
      
         May be implemented by derived classes.
      """
      return self.__class__.__name__
   
   def getDeps(self):
      """
         Returns the test dependencies
      """
      computedDeps = []
      for val in self.varsUsed.values():
         if type(val) is tuple:
            computedDeps.append(val[0])
         else:
            computedDeps.append(val)
      return self.deps + computedDeps

   def add(self):
      """
         Adds the test case to the testbed.
      """
      getTestbed().addTestCase(self)

   def __getitem__(self, key):
      # get the source testcase (and original name) for the variable
      try:
         sourceInfo = self.varsUsed[key]
      except KeyError, ex:
         raise BadTestCase('Variable "%s" is used but not defined in test '
                            'case %s' % (key, self.getId())
                           )
      
      if type(sourceInfo) is tuple:
         sourceName, sourceKey = sourceInfo
      else:
         sourceName = sourceInfo
         sourceKey = key
      try:
         return getTestbed()[sourceName, sourceKey]
      except KeyError, ex:
         raise BadTestCase('Variable "%s" (%s:%s) is used in test case '
                            '%s but is not defined' % 
                            (key, sourceName, sourceKey, self.getId())
                           )

   def __setitem__(self, key, val):
      if key not in self.varsDefined:
         raise BadTestCase('Variable "%s" is assigned but has not been '
                            'registered in test case %s' % (key, self.getId())
                           )
      getTestbed()[self.getId(), key] = val

class _MethodTestCase(TestCase):
   """
      Adapter class to allow us to use plain old python methods as test cases.
   """
   
   def __init__(self, method):
      self.__meth = method
      self.desc = method.__doc__
      if hasattr(method, 'deps'):
         self.deps = method.deps
      if hasattr(method, 'varsUsed'):
         self.varsUsed = method.varsUsed
      if hasattr(method, 'varsDefined'):
         self.varsDefined = method.varsDefined

   def run(self):
      return self.__meth()

   def getId(self):
      return self.__meth.__name__
   

class Testbed:
   
   def __init__(self):
      self.completedTestCases = {}
      self.testCases = {}
      self.testSeq = []
      self.__vars = {}
      self.__log = open('test.log', 'w')
   
   def addTestCase(self, testCase):
      """
         Adds a test case to the testbed - this is normally done by
         @TestCase.add()
         
         parms:
            testCase::
               [@TestCase]
      """
      self.testCases[testCase.getId()] = testCase
      self.testSeq.append(testCase)

   def __formatException():
      return ''.join(traceback.format_exception(*sys.exc_info()))
   __formatException = staticmethod(__formatException)

   def __doTest(self, testCase):
      # check all dependencies
      for dep in testCase.getDeps():
      
         # execute the dependency if we have not already
         if dep not in self.completedTestCases:
            
            # test the dependency and make sure that the test passed
            if not self.__doTest(self.testCases[dep]):
               return 0
         
         # make sure that the dependency passed
         elif not self.completedTestCases[dep]:
            return 0
         
      # do our test case
      try:
         print 'testing %s...' % testCase.getDesc(),
         self.log('=== testing %s - %s ===' % 
                   (testCase.getDesc(), testCase.getId())
                  )
         rc = testCase.run()
         self.log('* rc = %s' % repr(rc))
      except Exception, ex:
         self.log('* exception from %s:\n%s' % \
            (testCase.getId(), self.__formatException()))
         rc = 0
      except:
         self.log('* Exception not derived from Exception raised by %s:\n%s' %
                   (testCase.getId(), self.__formatException())
                  )
         rc = 0
      
      # record the results
      self.completedTestCases[testCase.getId()] = rc
      
      if not rc:
         print 'failed'
      else:
         print 'ok'
      return rc
   
   def runAll(self):
      """
         Runs all tests in the testbed.
      """
      allPassed = 1
      for testCase in self.testSeq:
         
         # check to see if we've already done it
         if self.completedTestCases.has_key(testCase.getId()):
            continue
         
         # check the dependency
         allPassed = allPassed and self.__doTest(testCase)
      
      if allPassed:
         print 'all tests passed'
      else:
         print 'some tests failed!!!'

   def addEZSuite(self, suite):
      """
         Adds an "EZ Test Suite": an instance of a class containing tests as
         methods, where the test description is the method doc string,
         and additional information is passed as method attributes.
         
         Method-tests are executed in alphabetical order.
         
         parms:
            suite::
               [instance] instance of a class in which all callable attributes
               whose names do not begin with underscores are test cases.
      """
      for attr in dir(suite):
   
         # exclude attributes beginning with underscores
         if attr[0] == '_':
            continue

         # check to see if the attribute value is callable, create a test
         # case for it if it is.
         val = getattr(suite, attr)
         if callable(val):
            _MethodTestCase(val).add()

   def __getitem__(self, (id, key)):
      return self.__vars[id, key]
   
   def __setitem__(self, (id, key), val):
      self.__vars[id, key] = val

   def has_key(self, (id, key)):
      return self.__vars.has_key((id, key))

   def log(self, msg):
      self.__log.write(msg + '\n')

_testbed = Testbed()
def getTestbed():
   """
      Returns the @Testbed singleton.
   """
   global _testbed
   return _testbed
         
if __name__ == '__main__':
   
   class TestCase1(TestCase):
      # test case description (required)
      desc = "test case run func"
      
      # indicates that TestCase1 is dependent on "Dependency".  This is
      # not really necessary since we use Dependency's "foo" variable.
      deps = ['Dependency']
      
      # this is a mapping of the variables used by TestCase1 to the test
      # cases that define them
      varsUsed = \
         {
         # we use the foo variable from "Dependency"
         'foo': 'Dependency',
         
         # we use the "bar" variable, which is an alias for the "foo" variable
         # from "VarSource".  Note that this creates an implicit dependency
         # on the VarSource test.
         'bar': ('VarSource', 'foo'),
         }
      
      def run(self):
         # if all of the variables have the correct values, the test
         # passes - note that we obtain the variable values through
         # a dictiary-like get from self.
         if self['bar'] == 'from var source' and self['foo'] == 1:
            return 1
         else:
            return 0

   # add the test case to the testbed
   TestCase1().add()

   class Dependency(TestCase):
      desc = 'dependency test case'
      
      # indicates that we define the variable "foo" in this test case
      varsDefined = ['foo']
      
      def run(self):
         # assign the variable (also through the dictionary type interface)
         self['foo'] = 1
         
         # as long as there were no exceptions, this test passes
         return 1

   Dependency().add()
         
   class VarSource(TestCase):
      desc = 'variable definitions'
      
      # this test case also defines "foo", the testbed knows that it is
      # "VarSource.foo" and not "Dependency.foo"
      varsDefined = ['foo']
      
      def run(self):
         self['foo'] = 'from var source'
         return 1

   VarSource().add()   

   class BunchOfTests:
      
      def a1(self):
         'test of method definitions'
         return True
      
      def a2(self):
         'raise an exception'
         raise Exception('this is an exception')
      
      def a3(self):
         'should not see this'
      a3.deps = [ 'a2' ]

   getTestbed().addEZSuite(BunchOfTests())      

   # run all tests.  Results (and exceptions) are logged to "test.log"
   getTestbed().runAll()


         
