#ifndef AXISBOX_H
#define AXISBOX_H

/**
 * Defines a region of three dimensional space that is orthongonal to the
 * axis.  As such, this region may be defined in terms of mininmum and maximum
 * x, y, and z values which are stored in the "min" and "max" public instance
 * variables.
 */
class AxisBox {
   public:
      Vector min, max;

      /** Default constructor - values of min and max are undefined. */
      AxisBox() {}

      AxisBox(const Vector &min, const Vector &max) : min(min), max(max) {}

      /** 
       * Constructs an infinite (if "convex" is true) or ninfinite (if
       * "convex" is false) AxisBox.  "ninfinite" is an inifinite box
       * inverted, so the minimum coordinates stretch to positive infinity
       * and the maximum coordinates stretch to negative infinity.  Ninfinate
       * boxes are useful because they become the minimum enclosing box of any
       * other box or point when expanded with the expandTo() method, making
       * them suitable for initial values of boxes constructed from a set of
       * points or boxes.
       */
      AxisBox(bool convex) {
         Float i = convex ? infinity : -infinity;
         min.x = min.y = min.z = -i;
         max.x = max.y = max.z = i;
      }

      /** expands the box to make it enclose the point */
      void expandTo(const Vector &point) {
         if (point.x < min.x)
            min.x = point.x;
         if (point.y < min.y)
            min.y = point.y;
         if (point.z < min.z)
            min.z = point.z;
         if (point.x > max.x)
            max.x = point.x;
         if (point.y > max.y)
            max.y = point.y;
         if (point.z > max.z)
            max.z = point.z;
      }

      /** expands the box to enclose "other" */
      void expandTo(const AxisBox &other) {
         if (other.min.x < min.x)
            min.x = other.min.x;
         if (other.min.y < min.y)
            min.y = other.min.y;
         if (other.min.z < min.z)
            min.z = other.min.z;
         if (other.max.x > max.x)
            max.x = other.max.x;
         if (other.max.y > max.y)
            max.y = other.max.y;
         if (other.max.z > max.z)
            max.z = other.max.z;
      }
};

inline std::ostream &operator <<(std::ostream &out, const AxisBox &box) {
   out << "AxisBox(" << box.min << ", " << box.max << ")";
   return out;
}

#endif
