#ifndef GAMEENGINE_H
#define GAMEENGINE_H

#include <sys/time.h>
#include <time.h>
#include <list>

#include <spug/Exception.h>

#include "Camera.h"
#include "Model.h"
#include "Moveable.h"
#include "PhysicalObject.h"

SPUG_EXCEPTION(FileDoesNotExistError);
SPUG_EXCEPTION(InvalidFileFormatError);

/**
 * Manages and schedules all game activities, including physics, networking
 * and user interaction.
 */
class GameEngine {

   private:
      // time of the last frame
      struct timeval lastEngineFrame, lastRenderFrame;

      // tracks seconds since the last frame of this type.
      Float engineDelta, renderDelta;

      // everybody's favority statistic
      Float renderFramesPerSecond;

      // the camera - the player's viewpoint and position
      Camera camera;

      // maximum distance we need to be able to see
      Float maxDistance;

      enum { framesPerSecond = 10 };

      // flags for the player "movement keys"
      bool forward, backward, left, right;

      Model map;

      // the player
      PhysicalObjectPtr player;

      // the list of game objects
      typedef std::list<GameObjectPtr> GameObjectSequence;
      GameObjectSequence objects;

      // the list of renderable objects
      typedef std::list<RenderableObjectPtr> RenderableObjectSequence;
      RenderableObjectSequence renderObjects;

   public:

      /** Default Constructor */
      GameEngine();

      /** 
       * Loads a new map from the given map file name.
       * @throws NonexistentFileError if the file does not exist
       * @throws InvalidFileFormatError if the file is not a valid map file.
       */
      void loadMap(const char *mapName);

      /**
       * Loads all game textures (this doesn't actually do anything.
       */
      void loadTextures();

      /**
       * Performs all activities in a standard frame.  This should be called
       * 10 times per second.  It is important that it be called as close to
       * schedule as possible, because it does physics processing - if not
       * called often enough, the game physics will slow down.  If called too
       * often, the game will have faster than real physics.
       */
      void processEngineFrame();

      /**
       * This should be called during every rendering cycle.  If more than a
       * tenth of a second has elapsed since the last frame, calls
       * processEngineFrame().
       */
      void processRenderFrame();

      /**
       * Renders the scene using opengl
       */
      void render();

      /**
       * This should be called during idle times.  Calls processRenderFrame()
       */
      void handleIdle();

      /** Returns the camera object - this is the player's viewpoint. */
      const Camera &getCamera() const {
	 return camera;
      }

      /** non-const method to return the camera object */
      Camera &getCamera() {
	 return camera;
      }

      /** 
       * Setter for "forward" mode.
       * If "forward" mode is set, the player will move forward.
       */
      void setForward(bool enabled) {
	 forward = enabled;
      }

      /** 
       * Getter for "forward" mode.
       * If "forward" mode is set, the player will move forward.
       */
      bool getForward() { return forward; }

      /** 
       * Sets backward mode.
       * "backward" mode indicates that the player is walking backward.
       */
      void setBackward(bool enabled) {
	 backward = enabled;
      }

      /** 
       * Gets backward mode.
       * "backward" mode indicates that the player is walking backward.
       */
      bool getBackward() { return backward; }

      /**
       * Getter for "left" mode.
       * If "left" mode is true, the player will strafe left.
       */
      void setLeft(bool enabled) {
	 left = enabled;
      }

      /**
       * Setter for "left" mode.
       * If "left" mode is true, the player will strafe left.
       */
      bool getLeft() { return left; }

      /**
       * Sets "right" mode, indicating that the player is strafing right.
       */
      void setRight(bool enabled) {
	 right = enabled;
      }

      /**
       * Sets "right" mode, indicating that the player is strafing right.
       */
      bool getRight() { return right; }

      /**
       * Returns the tiem in seconds since the last engine frame.  This 
       * will only be accurate during processing of an engine frame.
       */
      Float getEngineDelta() const { return engineDelta; }

      /**
       * Returns the tiem in seconds since the last render frame.  This 
       * will only be accurate during processing of an render frame.
       */
      Float getRenderDelta() const { return renderDelta; }

      /**
       * Adds a game object to the list of managed game objects.
       */
      void addGameObject(const GameObjectPtr &object);

      /**
       * Adds a renderable object to the list of objects that will be in the
       * engine's render loop.
       */
      void addRenderableObject(const RenderableObjectPtr &object);

      // XXX hack
      void makeCube(Model &cube);
};

#endif
