
#include "NativePersistor.h"
#include <spug/Reader.h>
#include <spug/FileReader.h>
#include "QuadObj.h"
#include "Model.h"
#include "PNG.h"
#include "Texture.h"
#include "Color.h"
#include "ColorMaterial.h"
#include "TextureMaterial.h"

#include <iostream>

using namespace spug;
using namespace std;

Tracer::Level NativePersistor::detailed = 
   Tracer::get().getLevel("NativePersistor.detailed");

Float NativePersistor::readFloat(Reader &reader) {
   Float data;
   if (reader.read(reinterpret_cast<Byte *>(&data), sizeof(Float)) != 
	sizeof(Float)
       )
      throw OutOfDataError("parsing Float");
   SPUG_TRACE(detailed, "readFloat() -> " << data);
   return data;
}

int NativePersistor::readInt(Reader &reader) {
   int data;
   if (reader.read(reinterpret_cast<Byte *>(&data), sizeof(int)) !=
	sizeof(int)
       )
      throw OutOfDataError("parsing int");
   SPUG_TRACE(detailed, "readInt() -> " << data);
   return data;
}

void NativePersistor::readString(Reader &reader, string &str) {
   size_t size = readInt(reader);
   Byte *buffer = new Byte[size];
   reader.read(buffer, size);

   str = string(reinterpret_cast<const char *>(buffer), size);
   delete buffer;
   SPUG_TRACE(detailed, "readString() -> " << str);
}

void NativePersistor::readColor(Reader &reader, Color &color) {
   color.red = readFloat(reader);
   color.green = readFloat(reader);
   color.blue = readFloat(reader);
   SPUG_TRACE(detailed, "read color");
}

void NativePersistor::readMaterial(Reader &reader, MaterialPtr &material) {
   // figure out what type of material this is...
   int materialType = readInt(reader);
   switch (materialType) {
      case 0: {
	 // color
	 Color color;
	 readColor(reader, color);
	 material = new ColorMaterial(color);
	 break;
      }
      case 1: {
	 // texture
	 std::string textureName;
	 readString(reader, textureName);
	 TexturePtr texture = new Texture(textureName.c_str());
	 Float xppus = readFloat(reader);
	 Float yppus = readFloat(reader);
	 material = new TextureMaterial(texture, xppus, yppus);
	 break;
      }
      default:
	 throw Exception("Unknown material type");
   }

   SPUG_TRACE(detailed, "read material");
}

void NativePersistor::readVector(Reader &reader, Vector &point) {
   point.x = readFloat(reader);
   point.y = readFloat(reader);
   point.z = readFloat(reader);
   SPUG_TRACE(detailed, "read point");
}

QuadObjPtr NativePersistor::readQuad(Reader &reader, 
				     const MaterialVector &materials) {

   // read the four points
   Vector c[4];
   for (int i = 0; i < 4; ++i)
      readVector(reader, c[i]);

   // read the color index
   int colorIndex = readInt(reader);

   SPUG_TRACE(detailed, "read quad");
   return new QuadObj(c[0], c[1], c[2], c[3], materials[colorIndex]);
}

void NativePersistor::readModel(Reader &reader, Model &model) {

   // read the colors
   int colorCount = readInt(reader);
   MaterialVector materials(colorCount);
   for (int i = 0; i < colorCount; ++i)
      readMaterial(reader, materials[i]);

   // read the polys
   int polyCount = readInt(reader);
   for (int i = 0; i < polyCount; ++i)
      model.addPoly(PolyObjPtr::ucast(readQuad(reader, materials)));

   SPUG_TRACE(detailed, "read model");
}
