#ifndef SCENE_H
#define SCENE_H

#include "RotMatrix.h"
/**
   A 3 dimensional scene.  This includes the origin (a Vector) and orientation
   (a RotMatrix) of the camera, and should include code to keep track of
   whether the scene is fully rendered.
 */
class Scene {

   private:
      Vector org;
      RotMatrix rot;
      Float maxDistance;

      // X factor and Y factor - we determine whether a point is visible by
      // multiplying z by these and checking whether the x & y coordinates lie
      // within result and -result.  A value of 1.0 corresponds to a 90 degree
      // field of view.
      static const Float xfact = 1.0, yfact = 1.0;

   public:
      // the direction vector - points directly forward.  Is equal to (0, 0,
      // -1) * rot
      Vector direction;

      Scene(const Vector &origin, const RotMatrix &orientation, 
            Float maxDistance
            );

      const RotMatrix &getRotMatrix() {
	 return rot;
      }

      const Vector &getOrigin() {
	 return org;
      }

      bool includes(const Vector &point) {

	 // immediately exclude points behind the camera
	 if (point.z < 0)
	    return false;

	 // calculate the maximum X and Y values for the given Z value
	 Float xmax = point.z * xfact, ymax = point.z * yfact;
	 return 
	    -xmax <= point.x <= xmax &&
	    -ymax <= point.y <= ymax;
      }

      /**
	 Assigns the perspective of the scene to the underlying rendering
	 engine.  This should be done for every scene.
       */
      void mapToCamera() const;
};
#endif
