
#include "Tri.h"

Vector Tri::normal(const Vector &coord0, const Vector &coord1, 
	           const Vector &coord2
		   ) {
   return (coord2 - coord0).crossProduct(coord1 - coord0);
}

// This is based on the algorithm for calculating intersection of a point and
// a triangle documented here:
//   http://softsurfer.com/Archive/algorithm_0105/algorithm_0105.htm#Segment-Plane
// the "point of intersection" code is largely ripped off from here:
//   http://nehe.gamedev.net/data/lessons/lesson.asp?lesson=30
// 
// The basic gist of this is 
//   1) calculate the vector where the ray interesects the plane defined by the
//   triangle
//   2) convert this vector to coordinates s & t which are in terms of the
//   edges of the triangle radiating from coord0 (so in terms of s & t,
//   coord0 = (s = 0, t = 0), coord1 = (1, 0), coord2 = (0, 1).
//   3) the intersection point is within the triangle if:
//      0 <= s
//      0 <= t
//      s + t <= 1

bool Tri::intersects(const Vector &org, const Vector &dir, 
		     Vector &intersection,
		     Float &distance
		     ) const {

   // get the "edge vectors" for the triangle
   Vector u = coord1 - coord0;
   Vector v = coord2 - coord0;

   // calculate the pointof intersection

   // construct the normal
   Vector normal = u.crossProduct(v);

   // calculate the dot-product
   Float dotProduct = dir * normal;

   // if the dot-product is zero, the ray is parallel to the plane and we will
   // get division by zero errors if we proceed.  We know that a parallel ray
   // won't intersect, so we can quit now.
   if (isZero(dotProduct))
      return false;

   // calculate the distance from the origin of the ray to the point of
   // intersection
   distance = (normal * (coord0 - org)) / dotProduct;

   // if distance is < 0, the intersection is _behind_ the origin in the given
   // direction, so we do not intersect
   if (distance < 0)
      return false;

   // finally, the point of interesection
   intersection = org + dir * distance;
   
   Vector w = intersection - coord0;

   // calculate the dot-products used for the determinations of s & v
   Float uv = u * v,
	 wv = w * v,
	 wu = w * u,
	 vv = v * v,
	 uu = u * u;

   // calculate s & t
   Float divisor = (uv * uv) - uu * vv;
   Float s = (uv * wv - vv * wu) / divisor;
   Float t = (uv * wu - uu * wv) / divisor;

   return s >= 0 && t >= 0 && s + t <= 1;
}

