#ifndef POINT_H
#define POINT_H

#include <iostream>
#include "basictypes.h"

/**
   An x,y,z coordinate triad.  Vectors are used to represent points in space,
   directions, offsets and anything else you can think of that can be
   described as x,y,z coordinates (colors are different, instead of x,y,z they
   have R,G,B)
 */
class Vector {
   public:
      Float x, y, z;

      Vector() : x(0), y(0), z(0) {}
      Vector(Float x, Float y, Float z) : x(x), y(y), z(z) {}

      friend std::ostream &operator <<(std::ostream &out, const Vector &p) {
	 out << "Vector(" << p.x << ", " << p.y << ", " << p.z << ")";
	 return out;
      }

      /**
	 Subtracts "other" from the vector, effectively making the vector
	 relative to "other".
       */
      Vector &operator -=(const Vector &other) {
	 x -= other.x;
	 y -= other.y;
	 z -= other.z;
	 return *this;
      }

      /**
	 Returns a new vector which is the vector with "other" subtracted,
	 effectively the vector relative to "other".
       */
      Vector operator -(const Vector &other) const {
	 return Vector(*this) -= other;
      }

      /**
	 Adds "other" to the vector, effectively making a vector
	 that was relative to "other" absolute.
       */
      Vector &operator +=(const Vector &other) {
	 x += other.x;
	 y += other.y;
	 z += other.z;
	 return *this;
      }

      /**
	 Returns a new vector which is the vector with "other" added,
	 effectively an absolute version of a vector relative to "other".
       */
      Vector operator +(const Vector &other) const {
	 return Vector(*this) += other;
      }

      /** Returns true if the two vectors are equal */
      bool operator ==(const Vector &other) const {
	 return x == other.x && y == other.y && z == other.z;
      }

      /** Returns the dot-product of the two vectors. */
      Float operator *(const Vector &other) const {
	 return x * other.x + y * other.y + z * other.z;
      }

      /** Returns the vector multiplied by the value */
      Vector operator *(Float val) const {
	 return Vector(*this) *= val;
      }

      /** Multiplies the vector by the value */
      Vector &operator *=(Float val) {
	 x *= val;
	 y *= val;
	 z *= val;
	 return *this;
      }

      /** returns the cross-product of two vectors */
      Vector crossProduct(const Vector &other) const {
	 return 
	    Vector(y * other.z - z * other.y,
		  z * other.x - x * other.z,
		  x * other.y - y * other.x
		  );
      }

      /** Returns the magnitude of the vector (distance from the origin) */
      Float size() const {
	 return sqrt(x * x + y * y + z * z);
      }

      /** 
       * Normalizes the vector.  Converts it to a vector in the same
       * direction but with a distance of 1.
       */
      void normalize() {
	 Float sz = size();
	 if (sz) {
	    x /= sz;
	    y /= sz;
	    z /= sz;
	 }
      }
};

#endif
