
#include "WurldSDLDriver.h"

#include <GL/gl.h>
#include <unistd.h>
#include "GameEngine.h"
#include "MouseHandler.h"
#include "Graphics.h"

WurldSDLDriver::WurldSDLDriver(GameEngine &engine, MouseHandler &mouseHandler) :
   engine(engine),
   mouseHandler(mouseHandler),
   terminated(false),
   initWidth(640),
   initHeight(480) {
}

void WurldSDLDriver::shutdown() {
   SDL_Quit();
   terminated = true;
}

void WurldSDLDriver::render() {
   glMatrixMode(GL_MODELVIEW);
   glClear(GL_COLOR_BUFFER_BIT);
   glClear(GL_DEPTH_BUFFER_BIT);
   engine.handleIdle();
   glFlush();

   SDL_GL_SwapBuffers();
}

void WurldSDLDriver::handleKeyPress(int keyCode) {
   switch (keyCode) {
      case 'q':
         shutdown();
	 break;
      case 'g':
	 // toggle mouse handler activation
	 if (mouseHandler.isActive())
	    mouseHandler.deactivate();
	 else
	    mouseHandler.activate(this);
	 break;
      case 'd':
	 engine.setForward(true);
	 break;
      case 'z':
	 engine.setLeft(true);
	 break;
      case 'c':
	 engine.setRight(true);
	 break;
      case 'x':
	 engine.setBackward(true);
	 break;
   }
}

void WurldSDLDriver::handleKeyRelease(int keyCode) {
   switch (keyCode) {
      case 'd':
	 engine.setForward(false);
	 break;
      case 'x':
	 engine.setBackward(false);
	 break;
      case 'z':
	 engine.setLeft(false);
	 break;
      case 'c':
	 engine.setRight(false);
	 break;
   }
}

int WurldSDLDriver::main(int argc, const char **argv) {

   SDL_Init(SDL_INIT_EVERYTHING);

   SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 8);
   SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 8);
   SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 8);

   surface = SDL_SetVideoMode(initWidth, initHeight, 24, 
                              SDL_FULLSCREEN | SDL_OPENGL);

   // other GL initialization
   Graphics::getInstance().initialize(initWidth, initHeight);

   // initialize the game engine
   engine.loadTextures();

   // let us do the event loop
   while (!terminated) {
      // render before event processing - if we terminate during event
      // processing, we don't want to render
      render();

      // process events
      SDL_Event event;
      while (SDL_PollEvent(&event)) {
         switch (event.type) {
            case SDL_KEYDOWN:
               handleKeyPress(event.key.keysym.sym);
               break;
            case SDL_KEYUP:
               handleKeyRelease(event.key.keysym.sym);
               break;
            case SDL_MOUSEMOTION:
               mouseHandler.handleMouseMove(event.motion.x, event.motion.y);
               break;
         }
      }
   }
}

void WurldSDLDriver::setDisplaySize(int width, int height) {
   initWidth = width;
   initHeight = height;
}

DisplayDriver::Size WurldSDLDriver::getSize() const {
   return DisplayDriver::Size(surface->w, surface->h);
}

void WurldSDLDriver::hideMouse() {
   SDL_ShowCursor(SDL_DISABLE);
}

void WurldSDLDriver::warpMouse(int x, int y) {
   SDL_WarpMouse(x, y);
}

void WurldSDLDriver::grabMouse(bool grab) {
   SDL_WM_GrabInput(grab ? SDL_GRAB_ON : SDL_GRAB_OFF);
}
