
#include <WurldGtkDriver.h>
#include <WurldSDLDriver.h>
#include "PythonPlugin.h"

#define MESSAGE(msg) std::cerr << msg << std::endl

/**
 * Returns the angle of rotation (in degrees) for a point defined by opposite
 * (the coordinate opposite the angle) and the adjacent (the coordinate
 * adjacent to the angle).  Assumes that this constitutes a normal vector (of
 * magnitude 1) so sqrt(opposite**2 + adjacent**2) == 1
 */
Float getAngleFromVector(Float opposite, Float adjacent) {
   if (adjacent < 0)
      return degrees(atan(opposite / adjacent));
   else if (opposite > 0)
      return -180 + degrees(atan(opposite / adjacent));
   else
      return 180 + degrees(atan(opposite / adjacent));
}

GameEngine engine;

// the mouse handler
MouseHandler mouseHandler(engine);

int main(int argc, const char **argv) {
   try {
      DisplayDriver *displayDriver;
      if (argc < 2 || strcmp(argv[1], "sdl"))
         displayDriver = new WurldGtkDriver(engine, mouseHandler);
      else
         displayDriver = new WurldSDLDriver(engine, mouseHandler);
      PythonPlugin plugin;
      plugin.setDisplayDriver(displayDriver);
      plugin.setGameEngine(&engine);
      try {
         plugin.handleInit();
      } catch (const PluginError &ex) {
         MESSAGE("Error initializing python plugin: " << ex);
         exit(1);
      }

      // load the initial map
      try {
         engine.loadMap("map.mdl");
      } catch (const InvalidFileFormatError &ex) {
         MESSAGE("Initial map has invalid file format: " << ex);
         exit(1);
      } catch (const FileDoesNotExistError &ex) {
         MESSAGE("Map file does not exist: " << ex);
         exit(1);
      }

      displayDriver->main(argc, argv);
      plugin.handleTerminate();
      delete displayDriver;
   } catch (const spug::Exception &ex) {
      MESSAGE("Received uncaught exception: " << ex);
      exit(1);
   } catch (const std::exception &ex) {
      MESSAGE("Received uncaught std::exception: " << ex.what());
      exit(1);
   } catch (...) {
      MESSAGE("Received uncaught anonymous exception");
      exit(1);
   }

}
