#==============================================================================
#
#  $Id$
#
"""
   A 3D, logo-like turtle.
"""
#
#  Copyright (C) 2007 Michael A. Muller
#
#  Permission is granted to use, modify and redistribute this code,
#  providing that the following conditions are met:
#
#  1) This copyright/licensing notice must remain intact.
#  2) If the code is modified and redistributed, the modifications must 
#  include documentation indicating that the code has been modified.
#  3) The author(s) of this code must be indemnified and held harmless
#  against any damage resulting from the use of this code.
#
#  This code comes with ABSOLUTELY NO WARRANTEE, not even the implied 
#  warrantee of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
#
#==============================================================================

import wurld
from util import vecToStr

vec = wurld.createVector

class PolyTool:
   """
      A PolyTool just creates polygons on a Model.  You specify the number
      of sides, every invocation of path() adds a new side to the current
      polygon until you have the requisite number of sides, at which point the
      polygon is stored and the next path will be added to a new polygon.
   """

   def __init__(self, material, sides = 4):
      self.model = wurld.createModel()
      self.__curPoly = []
      self.__sides = sides
      self.material = material

      if sides == 4:
	 self.__createPoly = wurld.createQuad

   def stamp(self, turtle):

      # add the current position to the list of points for the poly
      self.__curPoly.append(turtle.pos)

      # if we got the necessary number of sides, create the polygon
      if len(self.__curPoly) == self.__sides:
	 poly = self.__createPoly(*self.__curPoly + [self.material,])
	 self.model.addPoly(poly)
	 self.__curPoly = []

class ArrayTool:
   """
      This tool just captures the turtle's positions as an array of verteces.

      The array is available as the "array" public variable.
   """
   def __init__(self, array = None):
      if array is None:
         array = []
      self.array = array

   def stamp(self, turtle):
      self.array.append(turtle.pos)

class CrossSectionTool:

   """
      A CrossSectionTool allows you to draw a polygon that is a cross section
      of a tubular segment of a model.  These should generally be drawn
      perpendicular to the current direction.
   """

   def __init__(self, material, proc):
      """
         parms:
            material::
               [Material] the material to put on the surfaces
            proc::
               [callable<@Turtle>] a function that will move the turtle to
               draw the cross section.
      """
      self.material = material
      self.proc = proc
      self.__last = None
      self.model = wurld.createModel()

   def _next(self, array, index):
      index = (index + 1) % len(array)
      return array[index]

   def stamp(self, turtle):
      t2 = turtle.copy()
      array = []
      arrayTool = ArrayTool(array)
      t2.tool = arrayTool

      # get the verteces of the cross section
      self.proc(t2)

      if self.__last:
         for i in range(len(array)):
            print 'creating quad: %s %s %s %s' % \
               (vecToStr(self.__last[i]), 
                vecToStr(self._next(self.__last, i)),
                vecToStr(self._next(array, i)), 
                vecToStr(array[i])
                )
            quad = \
               wurld.createQuad(self.__last[i], self._next(self.__last, i),
                                self._next(array, i), 
                                array[i],
                                self.material
                                )

            self.model.addPoly(quad)

      self.__last = array

class Turtle:
   """
      A turtle is a graphics cursor.
      Public-vars:
	 tool::
	    The current tool.  The tool gets stamp()ed whenever the turtle
	    moves forward.
	 pos::
	    [Vector] the current position
	 rot::
	    [RotMatrix] the current direction\/orientation of the turtle
   """
   def __init__(self, tool = None, pos = vec(0, 0, 0), rot = None):
      self.pos = pos
      self.__primaryDir = vec(0, 0, -1)
      self.dir = vec(0, 0, -1)
      if rot is None:
	 rot = wurld.createYawRotMatrix(0)
      self.rot = rot
      self.tool = tool

   def __rotate(self, rotMatrix):
      # rotate the current rotation matrix
      oldRot = self.rot
      self.rot = wurld.createYawRotMatrix(0)
      self.rot.makeProduct(rotMatrix, oldRot)

      # rotate the direction vector
      self.rot.rotate(self.__primaryDir, self.dir)

   def yaw(self, degrees):
      self.__rotate(wurld.createYawRotMatrix(degrees))

   def roll(self, degrees):
      self.__rotate(wurld.createRollRotMatrix(degrees))

   def pitch(self, degrees):
      self.__rotate(wurld.createPitchRotMatrix(degrees))

   def stamp(self):
      """
	 Stamp the turtle onto the current tool.
      """
      self.tool.stamp(self)

   def forward(self, distance, stamp = True):
      start = self.pos
      self.pos = start + self.dir * distance
      if stamp and self.tool:
	 self.tool.stamp(self)

   def copy(self):
      """
	 Returns a copy of the turtle.  This is a shallow copy, but since the
	 turtle's methods do not mutate any of its objects, it should be safe
	 to use a shallow copy as long as the user doesn't mutate the
	 associated position or rotation objects.
      """
      return Turtle(self.tool, self.pos, self.rot)

def drawSquareXSect(turtle):
   turtle.pitch(-90)
   turtle.forward(1)
   turtle.yaw(-135)
   for i in range(3):
      turtle.forward(2)
      turtle.yaw(-90)

