/*===========================================================================*\

   $Id$

   This file describes the python bindings of the program.  It is used by the
   "cppre" program to generate pybind_wurld.cc.

   Copyright (C) 2006 Michael A. Muller

   Permission is granted to use, modify and redistribute this code,
   providing that the following conditions are met:
 
   1) This copyright/licensing notice must remain intact.
   2) If the code is modified and redistributed, the modifications must 
   include documentation indicating that the code has been modified.
   3) The author(s) of this code must be indemnified and held harmless
   against any damage resulting from the use of this code.

   This code comes with ABSOLUTELY NO WARRANTEE, not even the implied 
   warrantee of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

\*===========================================================================*/

module wurld;
include "PythonPluginInternal.h";

/**
 * The display driver singleton.  It generalizes the interface to the
 * underlying widget toolkit and window manager.
 */
@python_inst_pointer
class DisplayDriver {
   public:
      void setDisplaySize(int width, int height);
}

/**
 * Returns the global display driver.
 */
DisplayDriver getDisplayDriver();

/**
 * A game object.  These are added to the game engine for rendering and
 * callback procesing.
 */
@python_inst_rcptr
class GameObject { }

/**
 * A vector is used to represent a set of coordinates in 3D space.
 */
class Vector {
   public:
      /** The X, Y, and Z coordinates */
      /** @{ */
      @attr float x;
      @attr float y;
      @attr float z;
      /** @} */

      /** 
       * Multiplies the vector by the value, effectively changing the
       * magnitude of the vector.
       */
      Vector operator *(float val);

      /**
       * Returns the dot-product of the two vectors.
       */
      Vector operator +(Vector other);
}

/**
 * A rotation matrix is a 3x3 matrix that can be used to rotate a vector.  The
 * cross product of the original vector and the rotation matrix is the rotated
 * vector.
 *
 * To combine multiple rotations, create a cross-product using the
 * makeProduct() method.
 */
class RotMatrix { 
   public:

      /**
       * Rotates "point" and stores the rotated point in "result".
       */
      void rotate(Vector point, Vector result);

      /**
       * Converts the rotation matrix to the cross-product of a and b.
       *
       * This is used to combine multiple rotations - to rotate around
       * absolute axis, use a = current rotation, b = absolute rotation.  To
       * rotate around the axis relative to the current rotation, use a =
       * relative rotation, b = current rotation.
       */
      void makeProduct(RotMatrix a, RotMatrix b);
}

/**
 * Returns the rotation matrix for the given yaw angle (rotation around the Y
 * axis)
 */
RotMatrix createYawRotMatrix(float degrees);

/**
 * Returns the rotation matrix for the given pitch angle (rotation around the
 * X axis)
 */
RotMatrix createPitchRotMatrix(float degrees);

/**
 * Returns the rotation matrix for the given roll angle (rotation around the Z
 * axis.
 */
RotMatrix createRollRotMatrix(float degrees);

/**
 * Creates a vector.
 */
Vector createVector(float x, float y, float z);

/**
 * A texture object.  A texture is just an image used for constructing a
 * material.
 */
@python_inst_rcptr
class Texture { }

/** 
 * Creates a new texture from the given filename - this should be a PNG file 
 */
Texture createTexture(string filename);

/** A Material object.  Materials define the appearences of surfaces. */
@python_inst_rcptr
class Material { }

/** 
 * Creates a new material out of the given texture.
 */
Material createTextureMaterial(Texture texture, float xppus, float yppus);

/** A polygon (abstract class) */
@python_inst_rcptr
class PolyObj { }

/** Creates a quadrilateral */
PolyObj createQuad(Vector c0, Vector c1, Vector c2, Vector c3, 
		   Material material
		   );

/** A model.  Models are just collections of polygons. */
@python_inst_rcptr
class Model {
   public:
      /** Adds a new polygon to the model. */
      void addPoly(PolyObj poly);
}

/** Creates a new, empty model */
Model createModel();

/** Moveables are objects that can have models, position and velocity. */
@python_inst_rcptr
class Moveable {
   public:
      void setModel(Model model);

      void setPosition(Vector position);
      Vector getPosition();

      /** Get and set velocity - velocity is measured in units per second */
      /** @{ */
      void setVelocity(Vector velocity);
      Vector getVelocity();
      /** @} */
}

/** Create a new moveable object.  Position is (0, 0, 0), velocity is zero */
Moveable createMoveable();

/** 
 * A python object wraps a Moveable, allowing you to provide callback into
 * python code for every engine frame.  This is derived from GameObject,
 * although you wouldn't realize it from this prototype.
 */
@python_inst_rcptr
class PythonObject {
   public:
      /**
       * Sets the object to be called every engine frame (10 times per
       * second).  The callable should accept two arguments - the first will
       * be the game engine, the second will be the python object.
       */
      void setEngineFrameCallback(PyObject callback);

      /**
       * Sets the moveable object associated with the PythonObject.
       */
      void setMoveable(Moveable moveable);
}

/** returns a new python object */
PythonObject createPythonObject();

/**
 * The Game Engine.  This object manages all of the activity and rendering.
 */
@python_inst_pointer
class GameEngine {
   public:
      /**
       * Add a new game object
       */
      void addGameObject(GameObject object);
      void addRenderableObject(PythonObject object);
}

GameEngine getGameEngine();

@python_inst_pointer
class Graphics {
   public:
      void enableLighting(int enabled);
}

Graphics getGraphics();
